#pragma once

enum PGL_SPATIAL_STRUCTURE_TYPE
{
    PGL_SPATIAL_STRUCTURE_KDTREE = 0
};

enum PGL_DIRECTIONAL_DISTRIBUTION_TYPE
{
    PGL_DIRECTIONAL_DISTRIBUTION_PARALLAX_AWARE_VMM = 0,
    PGL_DIRECTIONAL_DISTRIBUTION_QUADTREE,
    PGL_DIRECTIONAL_DISTRIBUTION_VMM
};